#ifndef READSECTOR0_H_INCLUDED
#define READSECTOR0_H_INCLUDED

int readsector0 (struct checker *);
int readsector0_init (struct checker *);
void readsector0_free (struct checker *);

#endif /* READSECTOR0_H_INCLUDED */
